# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Frame": "api.ipynb",
         "validate_data_format": "api.ipynb",
         "read_data": "api.ipynb",
         "fcst_from_config": "api.ipynb",
         "perform_backtest": "api.ipynb",
         "parse_config": "api.ipynb",
         "setup_client": "api.ipynb",
         "app": "cli.ipynb",
         "run_forecast": "cli.ipynb",
         "date_features_dtypes": "core.ipynb",
         "GroupedArray": "core.ipynb",
         "simple_predict": "core.ipynb",
         "TimeSeries": "core.ipynb",
         "DataFreq": "data_model.ipynb",
         "DataFormat": "data_model.ipynb",
         "DataConfig": "data_model.ipynb",
         "BacktestConfig": "data_model.ipynb",
         "FeaturesConfig": "data_model.ipynb",
         "ForecastConfig": "data_model.ipynb",
         "ModelConfig": "data_model.ipynb",
         "LocalConfig": "data_model.ipynb",
         "ClusterConfig": "data_model.ipynb",
         "DistributedModelName": "data_model.ipynb",
         "DistributedModelConfig": "data_model.ipynb",
         "DistributedConfig": "data_model.ipynb",
         "FlowConfig": "data_model.ipynb",
         "DateFeatures": "data_model.ipynb",
         "Transforms": "data_model.ipynb",
         "DistributedTimeSeries": "distributed.core.ipynb",
         "DistributedForecast": "distributed.forecast.ipynb",
         "LGBMForecast": "distributed.models.lgb.ipynb",
         "XGBForecast": "distributed.models.xgb.ipynb",
         "Forecast": "forecast.ipynb",
         "generate_daily_series": "utils.ipynb",
         "data_indptr_from_sorted_df": "utils.ipynb",
         "backtest_splits": "utils.ipynb"}

modules = ["api.py",
           "cli.py",
           "core.py",
           "data_model.py",
           "distributed/core.py",
           "distributed/forecast.py",
           "distributed/models/lgb.py",
           "distributed/models/xgb.py",
           "forecast.py",
           "utils.py"]

doc_url = "https://Nixtla.github.io/mlforecast/"

git_url = "https://github.com/Nixtla/mlforecast/tree/main/"

def custom_doc_links(name): return None
